unit CoreExports;

interface

uses
  Windows;

type
  TDACEntry = packed record
    Channel: Word;
    Value: Word;
  end;
  TPDACEntry = ^TDACEntry;

  TDACMultiCallback = procedure (pDACData: TPDACEntry; DACDataCount: PLongWord); cdecl;

procedure InitDLL;

function HWave_Start(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
function HWave_Stop(DeviceIndex: LongWord): LongWord; cdecl;
function HWave_ResetClock(DeviceIndex: LongWord): LongWord; cdecl;
function HWave_ChangeClockRate(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
function HWave_RegisterCallback(DeviceIndex: Longword; pCallback: TDACMultiCallback): LongWord; cdecl;



implementation

uses
  System.SysUtils, System.Classes,
  AIOUSB;

function AnsiMessageBox(hWnd: HWND; Text, Caption: AnsiString; uType: UINT): Integer;
begin
  Result := MessageBoxA(hWnd, PAnsiChar(Text), PAnsiChar(Caption), uType);
  Text := '';
  Caption := '';
end;

const
  AUR_DAC_IMMEDIATE       = $B3;

type
  TWavePoint = packed record
    AFlags: Byte;
    AData: array[0..7] of Word;
    BFlags: Byte;
    BData: array[0..7] of Word;
  end;

  THWaveThread = class(TThread)
  public
    TicksPerSec, TicksPerPoint, NextPoint: Int64;
    iWavePointRead, iWavePointWrite, WavePointDepthLeft: Integer;
    bStarted: Boolean;
    NewTicksPerPoint: Int64;
    bHasNewTicksPerPoint: Boolean;

    constructor Create(NewDeviceIndex: LongWord; NewMilliseconds: Double);
    destructor Destroy; override;
  protected
    DeviceIndex: LongWord;

    procedure Execute; override;
  end;

var
  HWaveThread: array[0..31] of THWaveThread;
  pWaveCallback: array[0..31] of Pointer;
  hThreadTableMutex: THandle;

function WavePointFromDACEntries(pDACData: TPDACEntry; DACDataCount: LongWord): TWavePoint;
var
  pDE: TPDACEntry;
  pDB, pMB: PByte;
  I, J: Integer;

begin
  ZeroMemory(@Result, SizeOf(Result));
  pDB := PByte(@Result);
  pDE := pDACData;
  for I := 0 to DACDataCount - 1 do begin
    J := pDE.Channel and $F;
    pMB := pDB;
    Inc(pMB, J * 2 + (J div 8) + 1);
    PWord(pMB)^ := pDE.Value;

    pMB := pDB;
    Inc(pMB, (J div 8) * 17);
    pMB^ := pMB^ or (1 shl (J and 7));

    Inc(pDE);
  end;
end;



procedure DLLHandler(Reason: Integer);
var
  I: Integer;

begin
  case Reason of
    DLL_PROCESS_ATTACH: begin
      hThreadTableMutex := CreateMutex(nil, False, nil);
      for I := 0 to High(HWaveThread) do HWaveThread[I] := nil;
      for I := 0 to High(pWaveCallback) do pWaveCallback[I] := nil;
    end;
    DLL_PROCESS_DETACH: begin
      WaitForSingleObject(hThreadTableMutex, INFINITE);
        for I := 0 to High(HWaveThread) do
          FreeAndNil(HWaveThread[I])
        ;
      ReleaseMutex(hThreadTableMutex);
      CloseHandle(hThreadTableMutex);
      hThreadTableMutex := INVALID_HANDLE_VALUE;
    end;
  end;
end;

procedure InitDLL;
begin
  if PPointer(@DLLProc)^ = @DLLHandler then Exit;

  DLLProc := @DLLHandler;
  DLLHandler(DLL_PROCESS_ATTACH);
end;

function ValidateDI(var DI: LongWord): LongWord;
begin
  if DI < LongWord(Length(hWaveThread)) then begin
    Result := ERROR_SUCCESS;
    Exit;
  end;

  DI := ResolveDeviceIndex(DI);
  if DI >= LongWord(Length(hWaveThread)) then begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_INVALID_INDEX;
    Exit;
  end;

  Result := ERROR_SUCCESS;
end;

function HWave_Start(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
var
  NewThread: THWaveThread;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if Milliseconds <= 0 then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    NewThread := THWaveThread.Create(DeviceIndex, Milliseconds);

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if pWaveCallback[DeviceIndex] = nil then begin
        Result := ERROR_INVALID_FUNCTION;
      end
      else if HWaveThread[DeviceIndex] <> nil then begin
        Result := ERROR_SINGLE_INSTANCE_APP;
      end
      else begin
        Result := ERROR_SUCCESS;
        HWaveThread[DeviceIndex] := NewThread;
      end;
    ReleaseMutex(hThreadTableMutex);

    if Result = ERROR_SUCCESS then
      NewThread.Start
    else
      NewThread.Free
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_Stop(DeviceIndex: LongWord): LongWord; cdecl;
var
  OldThread: THWaveThread;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    OldThread := nil;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        Result := ERROR_SUCCESS;
        OldThread := HWaveThread[DeviceIndex];
        HWaveThread[DeviceIndex] := nil;
      end;
    ReleaseMutex(hThreadTableMutex);

    if OldThread <> nil then
      OldThread.Free
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_RegisterCallback(DeviceIndex: Longword; pCallback: TDACMultiCallback): LongWord; cdecl;
var
  pCallbackRaw: Pointer absolute pCallback;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] <> nil then begin
        Result := ERROR_BUSY;
      end
      else begin
        Result := ERROR_SUCCESS;
        pWaveCallback[DeviceIndex] := pCallbackRaw;
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_ResetClock(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        HWaveThread[DeviceIndex].bStarted := False;
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_ChangeClockRate(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
var
  NewTicksPerPoint: Int64;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    QueryPerformanceFrequency(NewTicksPerPoint);
    NewTicksPerPoint := Round(NewTicksPerPoint * Milliseconds / 1000);

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        HWaveThread[DeviceIndex].NewTicksPerPoint := NewTicksPerPoint;
        HWaveThread[DeviceIndex].bHasNewTicksPerPoint := True;
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

{
function HWave_QueueWavePoint(DeviceIndex: LongWord; const NewWavePoint: TWavePoint): LongWord;
begin
  try
    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        WaitForSingleObject(HWaveThread[DeviceIndex].hWavePointMutex, INFINITE);
          if ReleaseSemaphore(HWaveThread[DeviceIndex].hWavePointSem, 1, nil) then begin
            //It's safe to release the semaphore first because we hold the mutex.
            Result := ERROR_SUCCESS;
            HWaveThread[DeviceIndex].WavePointQueue[HWaveThread[DeviceIndex].iWavePointWrite] := NewWavePoint;
            HWaveThread[DeviceIndex].iWavePointWrite := (HWaveThread[DeviceIndex].iWavePointWrite + 1) mod HWaveThread[DeviceIndex].BufferDepth;
          end
          else begin
            Result := ERROR_NOT_READY;
          end;
        ReleaseMutex(HWaveThread[DeviceIndex].hWavePointMutex);
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_Queue(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
var
  NewWavePoint: TWavePoint;
  iDAC: Integer;
  pNextData: PWord;
  ProbeStr: AnsiString;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    try
      SetString(ProbeStr, PAnsiChar(pData), 16*SizeOf(pData^));
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    pNextData := pData;
    NewWavePoint.AFlags := $FF;
    NewWavePoint.BFlags := $FF;
    for iDAC := 0 to 7 do begin
      NewWavePoint.AData[iDAC] := pNextData^;
      Inc(pNextData);
    end;
    for iDAC := 0 to 7 do begin
      NewWavePoint.BData[iDAC] := pNextData^;
      Inc(pNextData);
    end;

    Result := HWave_QueueWavePoint(DeviceIndex, NewWavePoint);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_QueueMulti(DeviceIndex: LongWord; pDACData: TPDACEntry; DACDataCount: LongWord): LongWord; cdecl;
const
  ImmDACs = 16;

var
  ProbeStr: AnsiString;
  L: LongWord;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if DACDataCount = 0 then Exit; //If length is zero, then it's a nop.

    try
      SetString(ProbeStr, PAnsiChar(pDACData), DACDataCount*SizeOf(pDACData^));
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    pDE := pDACData;
    L := 0;
    for I := 0 to DACDataCount - 1 do begin
      if pDE.Channel > L then L := pDE.Channel;
      Inc(pDE);
    end;
    //L is now highest channel.

    if L >= ImmDACs then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    Result := HWave_QueueWavePoint(DeviceIndex, WavePointFromDACEntries(pDACData, DACDataCount));
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_FlushQueue(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else with HWaveThread[DeviceIndex] do begin
        WaitForSingleObject(hWavePointMutex, INFINITE);
          while WaitForSingleObject(hWavePointSem, 0) = WAIT_OBJECT_0 do begin
            Dec(iWavePointWrite);
            if iWavePointWrite < 0 then iWavePointWrite := iWavePointWrite + BufferDepth;
          end;
        ReleaseMutex(hWavePointMutex);
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;
}

{ THWaveThread }

constructor THWaveThread.Create(NewDeviceIndex: LongWord; NewMilliseconds: Double);
begin
  inherited Create(True);

  DeviceIndex := NewDeviceIndex;

  //hWavePointMutex := CreateMutex(nil, False, nil);

  QueryPerformanceFrequency(TicksPerSec);
  TicksPerPoint := Round(TicksPerSec * NewMilliseconds / 1000);
end;

destructor THWaveThread.Destroy;
begin
  Terminate;

  inherited;

  //CloseHandle(hWavePointMutex);
end;

procedure THWaveThread.Execute;
const
  BaseDACEntry: array[0..15] of TDACEntry = (
    (Channel: $0; Value: $4000), (Channel: $1; Value: $8000), (Channel: $2; Value: $8000), (Channel: $3; Value: $8000),
    (Channel: $4; Value: $8000), (Channel: $5; Value: $8000), (Channel: $6; Value: $8000), (Channel: $7; Value: $8000),
    (Channel: $8; Value: $8000), (Channel: $9; Value: $8000), (Channel: $A; Value: $8000), (Channel: $B; Value: $8000),
    (Channel: $C; Value: $8000), (Channel: $D; Value: $8000), (Channel: $E; Value: $8000), (Channel: $F; Value: $8000)
  );
var
  Naui: Int64;
  DACEntry: array[0..15] of TDACEntry;
  DACEntryCt: array of LongWord;
  ThisWavePoint: TWavePoint;
  pCallback: TDACMultiCallback;
  pCallbackRaw: Pointer absolute pCallback;

begin
  try
    bStarted := False;
    SetLength(DACEntryCt, 1); //Just to get it on the heap.
    SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);
    while not Terminated do begin
      WaitForSingleObject(hThreadTableMutex, INFINITE);
        pCallback := pWaveCallback[DeviceIndex];
      ReleaseMutex(hThreadTableMutex);
      if pCallbackRaw = nil then Break;

      DACEntryCt[0] := 16;
      Move(BaseDACEntry[0], DACEntry[0], SizeOf(BaseDACEntry));
      try
        pCallback(@DACEntry[0], @DACEntryCt[0]);
      except
        on E: Exception do begin
          SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);

          AnsiMessageBox(0, AnsiString(E.ClassName) + ' exception from callback:'#13#10 + AnsiString(E.Message), 'HWave', MB_ICONEXCLAMATION);
          Break;
        end;
      end;
      if Terminated then Break;

      ThisWavePoint := WavePointFromDACEntries(@DACEntry[0], DACEntryCt[0]);
      {
      WaitForSingleObject(hWavePointMutex, INFINITE);
        ThisWavePoint := WavePointQueue[iWavePointRead];
      ReleaseMutex(hWavePointMutex);
      iWavePointRead := (iWavePointRead + 1) mod BufferDepth;
      }

      if bStarted then begin
        repeat
          QueryPerformanceCounter(Naui);
          if (Naui - NextPoint) > 0 then Break;
          if Terminated then Break;

          if bHasNewTicksPerPoint then begin
            bHasNewTicksPerPoint := False;
            TicksPerPoint := NewTicksPerPoint;
          end;
        until False;
        if Terminated then Break;
      end
      else begin
        QueryPerformanceCounter(NextPoint);
        bStarted := True;
      end;
      Inc(NextPoint, TicksPerPoint);

      GenericVendorWrite(DeviceIndex, AUR_DAC_IMMEDIATE, 0, 0, SizeOf(ThisWavePoint), @ThisWavePoint);
    end;
  except
    on E: Exception do begin
      SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);

      AnsiMessageBox(0, AnsiString(E.ClassName) + ' exception from alternate thread context:'#13#10 + AnsiString(E.Message), 'HWave', MB_ICONEXCLAMATION);
    end;
  end;
end;

end.
